def list_dict_to_str(cookies_list_dict: list[dict[str, str]]) -> str:
    """将 cookies 的列表字典格式转换为字符串格式"""
    return "; ".join([c["name"] + "=" + c["value"] for c in cookies_list_dict])


def str_to_dict(cookies_str: str) -> dict[str, str]:
    """将 cookies 的字符串格式转换为字典格式"""
    return {c.split("=", maxsplit=1)[0].strip(): c.split("=", maxsplit=1)[-1].strip() for c in cookies_str.split(";")}


def dict_to_str(cookies_dict: dict[str, str]) -> str:
    """将 cookies 的字典格式转换为字符串格式"""
    return "; ".join([k + "=" + v for k, v in cookies_dict.items()])



import time
import random
import asyncio
import requests

from xspider.http.request import Request


class Downloader(object):
    def __init__(self) -> None:
        self._downloads: set[Request] | None = None

    def __len__(self) -> int:
        return len(self._downloads)

    def idle(self) -> bool:
        return len(self) == 0

    def open(self):
        self._downloads = set()

    async def download(self, request: Request):
        self._downloads.add(request)
        response = await self._download(request)
        self._downloads.remove(request)
        return response

    async def _download(self, request: Request):
        await asyncio.sleep(random.uniform(0, 1))
        return "<Response [200]>"



from asyncio import create_task, Semaphore

from xspider.type.types import Final, Task, Future, Coroutine


class Tasker(object):
    def __init__(self, concurrency: int = 8) -> None:
        self._concurrency: Final[int] = concurrency

        self._tasks: set[Task] | None = None
        self._semaphore: Semaphore | None = None

    def __len__(self) -> int:
        return len(self._tasks)

    def idle(self) -> bool:
        return len(self) == 0

    def open(self):
        self._tasks = set()
        self._semaphore = Semaphore(self._concurrency)

    async def create_task(self, coro: Coroutine) -> Task:
        task = create_task(coro)
        self._tasks.add(task)
        await self._semaphore.acquire()

        def done_callback(_fut: Future):
            self._tasks.remove(task)
            self._semaphore.release()

        task.add_done_callback(done_callback)
        return task



from . import configer
from . import crawler, crawler_process, downloader, engine, scheduler, tasker

__all__ = [
    "configer",
    "crawler", "crawler_process", "downloader", "engine", "scheduler", "tasker"
]



from asyncio import create_task

from xspider.item import Item
from xspider.spider import Spider
from xspider.http.request import Request
from xspider.core.scheduler import Scheduler
from xspider.core.downloader import Downloader
from xspider.core.tasker import Tasker
from xspider.utils.project import transform
from xspider.type.types import Iterator, Callable
from xspider.type.validators import is_coroutine
from xspider.exceptions import OutputException


class Engine(object):
    def __init__(self):
        self.scheduler: Scheduler | None = None
        self.downloader: Downloader | None = None
        self.tasker: Tasker | None = None

        self.spider: Spider | None = None
        self.start_requests: Iterator[str] | None = None
        self.running: bool = False

    async def start_spider(self, spider: Spider):
        self.scheduler = Scheduler()
        if hasattr(self.scheduler, "open"):
            self.scheduler.open()
        self.downloader = Downloader()
        if hasattr(self.downloader, "open"):
            self.downloader.open()
        self.tasker = Tasker()
        if hasattr(self.tasker, "open"):
            self.tasker.open()

        self.spider = spider
        self.start_requests = iter(spider.start_requests())
        self.running = True

        await self._open_spider()

    async def _open_spider(self):
        crawling = create_task(self.crawl())
        await crawling

    async def crawl(self):
        while self.running:
            if (request := await self._get_dequeue_request()) is not None:
                await self._crawl(request)
            else:
                try:
                    start_request = next(self.start_requests)  # noqa
                except StopIteration:
                    self.start_requests = None
                except Exception as e:
                    if not await self._exit():
                        continue
                    self.running = False
                else:
                    await self.enqueue_request(start_request)

    async def enqueue_request(self, request):
        await self._schedule_request(request)

    async def _schedule_request(self, request):
        # todo: 处理去重
        await self.scheduler.enqueue_request(request)

    async def _get_dequeue_request(self):
        return await self.scheduler.dequeue_request()

    async def _crawl(self, request):
        async def crawl_task():
            outputs = await self._fetch(request)
            if outputs:
                await self._handle_spider_outputs(outputs)

        await self.tasker.create_task(crawl_task())

    async def _fetch(self, request: Request):
        async def success(response):
            callback: Callable = request.callback or self.spider.parse

            # type(callback(response))
            # <class 'generator'>
            # <class 'async_generator'>
            # <class 'NoneType'>

            if outputs := callback(response):
                if is_coroutine(outputs):
                    await outputs
                else:
                    return transform(outputs)

        resp = await self.downloader.download(request)
        outs = await success(resp)
        return outs

    async def _handle_spider_outputs(self, outputs):
        async for output in outputs:
            if isinstance(output, Request):
                await self.enqueue_request(output)
            elif isinstance(output, Item):
                pass
            else:
                raise OutputException(f"{type(self.spider)} 必须返回 'Request' 或 'Item'！")

    async def _exit(self):
        if self.scheduler.idle() and self.downloader.idle() and self.tasker.idle():
            return True
        return False



import re


def camel_to_snake(text: str) -> str:
    """
    >>> camel_to_snake('CamelCaseString')
    'camel_case_string'

    :param text:
    :return:
    """
    text = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    text = re.sub("([a-z0-9])([A-Z])", r"\1_\2", text).lower()
    return text


def snake_to_camel(text: str) -> str:
    """
    >>> snake_to_camel('snake_case_string')
    'SnakeCaseString'

    :param text:
    :return:
    """
    return "".join(word.capitalize() for word in text.split("_"))



from ujson import loads
from pprint import pformat
from importlib import import_module

from xspider.type.types import Final, Any, ModuleType, Iterator, ConfigerItemsView, ConfigerKeysView, \
    ConfigerValuesView, MutableMapping
from xspider.type.validators import is_str
from xspider.core.configer import default_config
from xspider.exceptions import ConfigerException


class Configer(MutableMapping):

    def __init__(self, config: dict[str:Any] | None = None) -> None:
        self._config: Final[dict[str, Any]] = dict()
        self.set_config(default_config)
        self.update_config(config)

    def __len__(self) -> int:
        return len(self._config)

    def __iter__(self) -> Iterator:
        return iter(self._config)

    def __contains__(self, key: str) -> bool:
        return key in self._config

    def __getitem__(self, key: str) -> Any | None:
        if key not in self:
            return None
        return self._config[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.set(key, value)

    def __delitem__(self, key: str) -> None:
        self.delete(key)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} _config={pformat(self._config)}>"

    __repr__ = __str__

    def get(self, key: str, default: Any | None = None) -> Any | None:
        return self[key] if key in self else default

    def get_str(self, key: str, default: str = "") -> str:
        return str(self.get(key, default))

    def get_int(self, key: str, default: int = 0) -> int:
        return int(self.get(key, default))

    def get_float(self, key: str, default: float = 0.0) -> float:
        return float(self.get(key, default))

    def get_bool(self, key: str, default: bool = False) -> bool:
        value = self.get(key, default)
        try:
            return bool(int(value))
        except ValueError:
            if value in ("True", "true", "TRUE"):
                return True
            if value in ("False", "false", "FALSE"):
                return False
            raise ConfigerException(f"布尔类型配置的合法取值应为：0 或 1，True 或 False，'0' 或 '1'，"
                                    f"'True' 或 'False'，'true' 或 'false'，'TRUE' 或 'FALSE'！")

    def get_list(self, key: str, default: list[Any] | None = None) -> list[Any]:
        value = self.get(key, default if default is not None else [])
        if isinstance(value, str):
            value = value.split(",")
        return list(value)

    def get_dict(self, key: str, default: dict[Any, Any] = None) -> dict[Any, Any]:
        value = self.get(key, default if default is not None else {})
        if isinstance(value, str):
            value = loads(value)
        return dict(value)

    def set(self, key: str, value: Any) -> None:
        self._config[key] = value

    def delete(self, key: str) -> None:
        del self._config[key]

    def set_config(self, module: str | ModuleType) -> None:
        if isinstance(module, str):
            module = import_module(module)
        for key in dir(module):
            if is_str(key) and key.isupper():
                value = getattr(module, key)
                self.set(key, value)

    def update_config(self, config: dict[str, Any] | None = None) -> None:
        if config is not None:
            for key, value in config.items():
                self.set(key, value)

    def items(self) -> ConfigerItemsView:
        return ConfigerItemsView(self)

    def keys(self) -> ConfigerKeysView:
        return ConfigerKeysView(self)

    def values(self) -> ConfigerValuesView:
        return ConfigerValuesView(self)


__all__ = [
    "Configer",
    "default_config"
]



from inspect import isgenerator, isasyncgen, iscoroutine

from xspider.type.types import Any, TypeVar, TypeGuard


def is_int(obj: Any) -> TypeGuard[int]:
    return isinstance(obj, int) and not isinstance(obj, bool)


def is_str(val: Any) -> TypeGuard[str]:
    return isinstance(val, str)


def is_list(val: Any) -> TypeGuard[list[Any]]:
    return isinstance(val, list)


T = TypeVar("T")


def is_list_of(val: list[Any], type_: type[T]) -> TypeGuard[list[T]]:
    return isinstance(val, list) and all(isinstance(i, type_) for i in val)


is_generator = isgenerator
is_async_generator = isasyncgen
is_coroutine = iscoroutine

__all__ = [
    "is_str",
    "is_list_of",
    "is_generator", "is_async_generator", "is_coroutine"
]



from . import core, http, item, spider, type, utils
from . import exceptions
from .spider import Spider
from .http.request import Request
from .utils.project import get_configer
from .type import types, validators

__all__ = [
    "core", "http", "item", "spider", "type", "utils",
    "exceptions",
    "Spider", "Request", "get_configer", "types", "validators"
]



from xspider.utils.spider_priority_queue import SpiderPriorityQueue
from xspider.http.request import Request


class Scheduler(object):
    def __init__(self) -> None:
        self._schedules: SpiderPriorityQueue | None = None

    def __len__(self) -> int:
        return self._schedules.qsize()

    def idle(self) -> bool:
        return len(self) == 0

    def open(self) -> None:
        self._schedules = SpiderPriorityQueue()

    async def enqueue_request(self, request: Request) -> None:
        await self._schedules.put(request)

    async def dequeue_request(self) -> Request:
        return await self._schedules.get()



import arrow


def now(fmt: str = "YYYY-MM-DD HH:mm:ss") -> str:
    return arrow.now().format(fmt)


def delta(start_datetime_str: str, end_datetime_str: str) -> int:
    start_datetime = arrow.get(start_datetime_str).datetime
    end_datetime = arrow.get(end_datetime_str).datetime
    total_seconds = (end_datetime - start_datetime).total_seconds()
    return int(total_seconds)



from htmlmin import minify
from html import escape as html_escape, unescape as html_unescape

from xspider.type.types import Any


def compress(html: str, /, **kwargs: Any) -> str:
    return minify(html, **kwargs)


def escape(text: str) -> str:
    """
    >>> escape("<")
    '&lt;'

    """
    return html_escape(text)


def unescape(text: str) -> str:
    """
    >>> unescape("&lt;")
    '<'

    """
    return html_unescape(text)



import hashlib

from xspider.exceptions import TransformException
from xspider.type.types import Literal, Any, ModuleType, Generator, AsyncGenerator, TYPE_CHECKING
from xspider.type.validators import is_generator, is_async_generator

if TYPE_CHECKING:
    from xspider.core.configer import Configer


async def transform(result: Generator | AsyncGenerator):
    if is_generator(result):
        for i in result:
            yield i

    elif is_async_generator(result):
        async for i in result:
            yield i

    else:
        raise TransformException("callback 返回值类型必须是 'generator' 或 'async_generator'！")


def get_configer(config: str | ModuleType = "config") -> "Configer":
    from xspider.core.configer import Configer

    configer = Configer()
    configer.set_config(config)
    return configer


ALGO_TYPE = Literal[
    "md5", "sha1", "sha224", "sha256", "sha384", "sha512", "blake2b", "blake2s", "sha3_224", "sha3_256", "sha3_384",
    "sha3_512", "shake_128", "shake_256"
]


def gen_data_id(
        *args: Any,
        keys: list | None = None, item: dict | None = None,
        algo_type: ALGO_TYPE = "sha256"
) -> str:
    """
    >>> gen_data_id('123456')
    '8d969eef6ecad3c29a3a629280e686cf0c3f5d5a86aff3ca12020c923adc6c92'

    :param args:
    :param keys:
    :param item:
    :param algo_type:
    :return:
    """
    m = hashlib.new(algo_type)
    if args:
        values = args
    elif keys is not None and item is not None:
        if isinstance(keys, list) and isinstance(item, dict):
            values = [item[k] for k in keys if k in item]
        else:
            raise ValueError(f"keys 必须是列表，item 必须是字典！")
    elif item is not None:
        values = [item[k] for k in sorted(item.keys())]
    else:
        raise ValueError(f"args 或 keys 和 item 或 item 必须提供之一！")

    data = list(map(lambda x: str(x), values))

    for i in data:
        m.update(i.encode())

    data_id = m.hexdigest()
    return data_id



from xspider.type.types import Literal, Any, Self, Callable


class Request(object):
    def __init__(
            self,
            url: str,
            /,
            *,
            callback: Callable[..., Any] | None = None,
            method: Literal["GET", "POST", "PUT", "DELETE"] = "GET",
            headers: dict[str, str] | None = None,
            body: bytes | str | None = None,
            cookies: dict[str, str] | None = None,
            priority: int = 0,
            proxy: dict[str, str] | None = None
    ):
        self.url = url
        self.callback = callback
        self.method = method
        self.headers = headers
        self.body = body
        self.cookies = cookies
        self.priority = priority
        self.proxy = proxy

    def __lt__(self, other: Self) -> bool:
        return self.priority < other.priority

    def __repr__(self) -> str:
        return f"<Request {self.method} {self.url}>"


__all__ = [
    "Request"
]



from asyncio import PriorityQueue, wait_for, TimeoutError

from xspider.type.types import Any


class SpiderPriorityQueue(PriorityQueue):
    def __init__(self, maxsize: int = 0):
        super(SpiderPriorityQueue, self).__init__(maxsize=maxsize)

    async def get(self) -> Any | None:
        fut = super().get()
        try:
            return await wait_for(fut, timeout=0.1)
        except TimeoutError:
            return None



from xspider.http.request import Request
from xspider.type.types import Generator
from xspider.type.validators import is_str, is_list_of


class Spider(object):
    start_url: str | None = None
    start_urls: list[str] | None = None

    def __init__(self) -> None:
        pass

    def start_requests(self) -> Generator[Request, None, None]:
        if is_str(url := self.start_url):
            yield Request(url)
        if is_list_of(urls := self.start_urls, str):
            for url in urls:
                yield Request(url)

    def parse(self, response):
        raise NotImplementedError("必须实现 parse 方法！")


__all__ = [
    "Spider"
]



from types import ModuleType
from asyncio import Task, Future
from typing import (
    Literal,
    Final,
    Any,
    Self,
    TypeVar,
    TypeGuard,
    TypeAlias, Never, NoReturn, TYPE_CHECKING
)
from collections.abc import (
    Iterator, Callable, Generator, AsyncGenerator, AsyncIterable, AsyncIterator, Coroutine, MutableMapping,
    ItemsView, KeysView, ValuesView
)

from xspider.utils.string import camel_to_snake


class ConfigerItemsView(ItemsView):
    def __repr__(self) -> str:
        return f"{camel_to_snake(self.__class__.__name__)}({list(self)})"


class ConfigerKeysView(KeysView):
    def __repr__(self) -> str:
        return f"{camel_to_snake(self.__class__.__name__)}({list(self)})"


class ConfigerValuesView(ValuesView):
    def __repr__(self) -> str:
        return f"{camel_to_snake(self.__class__.__name__)}({list(self)})"


__all__ = [
    "ModuleType",

    "Task", "Future",

    "Literal", "Final", "Any", "Self", "TypeVar", "TypeGuard", "TYPE_CHECKING",

    "Iterator", "Callable", "Generator", "AsyncGenerator", "Coroutine", "MutableMapping", "ConfigerItemsView",
    "ConfigerKeysView", "ConfigerValuesView"
]
